﻿import sys
import clr
import re
import System
clr.AddReferenceToFile("DotPUniGenEngine.dll")
from tfm.DotPUniGen.Engine import *
from tfm.Common import *

sN = System.Environment.NewLine
lDefault = Language.English

def Info():
	return """DotP2014 Output
v1.1
by thefiremind
Generates a DotP2014-compatible XML string from card information"""

def Purpose():
	return PluginPurpose.Output

def Get(crdCard, sPrefix):
	sXml = '<?xml version="1.0"?>' + sN + '<CARD_V2 ExportVersion="1">' + sN
	# compute missing information
	sCardName = StringUtils.Nameify(crdCard.GetTitle(Language.English))
	sPrefixedId = crdCard.MultiverseID
	if len(sPrefix) > 0:
		while len(sPrefixedId) < 6:
			sPrefixedId = '0' + sPrefixedId
		sPrefixedId = sPrefix + sPrefixedId
	crdCard.FileName = sCardName + '_' + sPrefixedId
	# add filename
	sXml += SimpleNode('FILENAME', 'text', crdCard.FileName)
	# add cardname
	sXml += SimpleNode('CARDNAME', 'text', sCardName)
	# add title
	sXml += ('  <TITLE>' + sN)
	for lLang in MiscUtils.GetValues[Language]():
		sTitle = crdCard.GetTitle(lLang)
		if len(sTitle) < 1:
			sTitle = crdCard.GetTitle(lDefault)
		sXml += LocalizedTextNode(lLang, sTitle)
	sXml += ('  </TITLE>' + sN)
	# add prefixed multiverse id
	sXml += SimpleNode('MULTIVERSEID', 'value', sPrefixedId)
	# add prefixed multiverse id as art id as well
	sXml += SimpleNode('ARTID', 'value', sPrefixedId)
	# add colors if necessary
	if crdCard.ForcedColors:
		sColors = ""
		for ccColor in MiscUtils.GetValues[CardColor]():
			if ccColor != CardColor.Colorless and crdCard.HasColor(ccColor):
				sColors += Constants.ColorToCode(ccColor)
		if len(sColors) < 1:
			sColors = "C"
		sXml += SimpleNode('COLOUR', 'value', sColors)
	# add artist
	sXml += SimpleNode('ARTIST', 'name', crdCard.Artist.replace("&", "&amp;"))
	# add casting cost
	sXml += SimpleNode('CASTING_COST', 'cost', crdCard.CastingCost)
	# add flavor if present
	sDefaultFlavor = crdCard.GetFlavor(lDefault)
	if len(sDefaultFlavor) > 0:
		sXml += ('  <FLAVOURTEXT>' + sN)
		for lLang in MiscUtils.GetValues[Language]():
			sFlavor = crdCard.GetFlavor(lLang)
			if len(sFlavor) < 1:
				sFlavor = sDefaultFlavor
			sXml += LocalizedTextNode(lLang, sFlavor)
		sXml += ('  </FLAVOURTEXT>' + sN)
	# add supertypes
	for sSupertype in crdCard.GetSupertypesArray():
		sXml += SimpleNode('SUPERTYPE', 'metaname', sSupertype)
	# add types
	for sType in crdCard.GetTypesArray():
		sXml += SimpleNode('TYPE', 'metaname', sType)
	# add sub-types
	for sSubType in crdCard.GetSubTypesArray():
		sXml += SimpleNode('SUB_TYPE', 'metaname', sSubType)
	# add expansion
	sXml += SimpleNode('EXPANSION', 'value', crdCard.Expansion)
	# add rarity
	sRarity = crdCard.GetRarity().ToString()[:1]
	if sRarity == "S":
		# special rarity would crash the game, so we change it into something that works but is also easy to notice
		sRarity = "T"
	sXml += SimpleNode('RARITY', 'metaname', sRarity)
	# add P/T if present
	bGotPower = False
	if len(crdCard.Power) > 0:
		bGotPower = True
		sXml += SimpleNode('POWER', 'value', crdCard.Power)
		sXml += SimpleNode('TOUGHNESS', 'value', crdCard.Toughness)
	sXml += AbilitiesAndHelp(crdCard)
	# add the default combat SFX nodes if P/T were added before
	if bGotPower:
		sXml += ('  <SFX text="COMBAT_BLUNT_LARGE_ATTACK" power_boundary_min="4" power_boundary_max="-1" />' + sN)
		sXml += ('  <SFX text="COMBAT_BLUNT_SMALL_ATTACK" power_boundary_min="1" power_boundary_max="3" />' + sN)
	sXml += '</CARD_V2>'
	return sXml

def AbilitiesAndHelp(crdCard):
	sXml = ""
	lstAbilityNodes = []
	lstHelpNodes = []
	for iIndex in range(0, crdCard.GetAbilitiesCount()):
		tupAH = ProcessAbility(crdCard, iIndex)
		lstAbilityNodes.append(tupAH[0])
		if len(tupAH[1]) > 0:
			lstHelpNodes.append(tupAH[1])
	for sAbility in lstAbilityNodes:
		sXml += sAbility
	for sHelp in lstHelpNodes:
		sXml += sHelp
	return sXml

def ProcessAbility(crdCard, iIndex):
	sAbility = ""
	sHelp = ""
	sEnglishText = crdCard.GetAbilityText(iIndex, Language.English)
	# search for simple templates
	for sPattern, sTemplate in dicSimpleTemplates.iteritems():
		moMatch = re.search('^' + sPattern + '$', sEnglishText)
		if moMatch:
			if sPattern in dicSimpleTemplatesHelp.keys():
				sHelp = '  ' + dicSimpleTemplatesHelp[sPattern] + sN
			return '  ' + sTemplate.replace('\n', sN) + sN, sHelp
	sAbilityType = ""
	# at this point, localized text needs to be extracted for sure
	sLocText = ""
	for lLang in MiscUtils.GetValues[Language]():
		sText = crdCard.GetAbilityText(iIndex, lLang)
		if len(sText) < 1:
			sText = crdCard.GetAbilityText(iIndex, lDefault)
		sLocText += LocalizedTextNode(lLang, sText)
	# search for header templates
	sHeader = ""
	for sPattern, sTemplate in dicPatternHeaders.iteritems():
		moMatch = re.search(sPattern, sEnglishText)
		if moMatch:
			if sPattern in dicPatternHeadersHelp.keys():
				sHelp = '  ' + dicPatternHeadersHelp[sPattern] + sN
			# the ability type will be needed later
			sAbilityType = sTemplate[sTemplate.find('<')+1:sTemplate.find('_')]
			sHeader = '  ' + sTemplate + sN
			break
	if len(sHeader) < 1:
		# try and guess ability type
		for sPattern, sTemplate in dicAbilityTypeTemplates.iteritems():
			moMatch = re.search(sPattern, sEnglishText)
			if moMatch:
				sAbilityType = sTemplate
				break
		if len(sAbilityType) < 1:
			if crdCard.HasType(CardType.Instant) or crdCard.HasType(CardType.Sorcery):
				sAbilityType = 'SPELL'
			else:
				sAbilityType = 'STATIC'
		sHeader = '  <' + sAbilityType + '_ABILITY>' + sN
	sAbility += sHeader
	sAbility += sLocText
	# in case of activated or utility ability, search for cost templates
	if sAbilityType in ['ACTIVATED', 'UTILITY']:
		sCost = ""
		for sPattern, sTemplate in dicPatternCosts.iteritems():
			moMatch = re.search(sPattern, sEnglishText)
			if moMatch:
				for iIndex in range(1, len(moMatch.groups())):
					sTemplate = sTemplate.replace('@' + str(iIndex) + '@', str(moMatch.group(iIndex)))
				sCost += ('    ' + sTemplate.replace('\n', sN) + sN)
		sAbility += sCost
	# in case of triggered ability, search for trigger templates
	elif sAbilityType == 'TRIGGERED':
		for sPattern, sTemplate in dicPatternTriggers.iteritems():
			moMatch = re.search(sPattern, sEnglishText)
			if moMatch:
				for iIndex in range(1, len(moMatch.groups())):
					sTemplate = sTemplate.replace('@' + str(iIndex) + '@', str(moMatch.group(iIndex)))
				sAbility += ('    ' + sTemplate.replace('\n', sN) + sN)
				break
	# search for effect templates
	for sPattern, sTemplate in dicPatternEffects.iteritems():
		moMatch = re.search(sPattern, sEnglishText)
		if moMatch:
			if sPattern in dicPatternEffectsHelp.keys():
				sHelp = '  ' + dicPatternEffectsHelp[sPattern] + sN
			for iIndex in range(1, len(moMatch.groups())):
				sTemplate = sTemplate.replace('@' + str(iIndex) + '@', str(moMatch.group(iIndex)))
			sAbility += ('    ' + sTemplate.replace('\n', sN) + sN)
			break
	sAbility += NodeEnd(sAbility)
	return sAbility, sHelp

def SimpleNode(sNodeName, sAttributeName, sAttributeValue):
	return '  <' + sNodeName + ' ' + sAttributeName + '="' + sAttributeValue + '" />' + sN

def LocalizedTextNode(lLanguage, sContents):
	return '    <LOCALISED_TEXT LanguageCode="' + Constants.LanguageToCode(lLanguage) + '"><![CDATA[' + sContents + ']]></LOCALISED_TEXT>' + sN

def NodeEnd(sNode):
	moMatch = re.search(r'^(\s*)<(\w+)(?=[> ])', sNode)
	if moMatch:
		return moMatch.group(1) + '</' + moMatch.group(2) + '>' + sN
	else:
		return ""

def SimpleCostPattern():
	sPattern = r'((\{(\d+|['
	for ceElem in MiscUtils.GetValues[CostElement]():
		if not ceElem in [CostElement.Tap, CostElement.Untap, CostElement.Planeswalk, CostElement.Chaos]:
			sPattern += re.escape(Constants.CostElementToTextString(ceElem))
	sPattern += r'])\})+)'
	return sPattern

def Test():
	crdDummy = Card()
	print Get(crdDummy, '100')

#--------------------------------------------------------------------------------------------
# SIMPLE TEMPLATES
# The following templates are placed in the card when the exact ability name has been found.
#--------------------------------------------------------------------------------------------
dicSimpleTemplates = dict([

	("First strike",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[First strike]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Initiative]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Daña primero.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Erstschlag]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Attacco improvviso]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[先制攻撃]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[선제공격]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Первый удар]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Iniciativa]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_FIRST_STRIKE" />
  </STATIC_ABILITY>"""),

	("Trample",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Trample]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Piétinement]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Arrolla.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Verursacht Trampelschaden]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Travolgere]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[トランプル]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[돌진]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Пробивной удар]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Atropelar]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_TRAMPLE" />
  </STATIC_ABILITY>"""),

	("Double strike",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Double strike]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Double initiative]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Daña dos veces.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Doppelschlag]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Doppio attacco]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[二段攻撃]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[이단공격]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Двойной удар]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Golpe duplo]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_DOUBLE_STRIKE" />
  </STATIC_ABILITY>"""),

	("Deathtouch",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Deathtouch]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Contact mortel]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Toque mortal.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Todesberührung]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Tocco letale]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[接死]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[치명타]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Смертельное касание]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Toque mortífero]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_DEATHTOUCH" />
  </STATIC_ABILITY>"""),

	("Lifelink",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Lifelink]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Lien de vie]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Vínculo vital.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Lebensverknüpfung]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Legame vitale]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[絆魂]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[생명연결]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Цепь жизни]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Vínculo com a vida]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_LIFELINK" />
  </STATIC_ABILITY>"""),

	("Reach",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Reach]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Portée]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Alcance.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Reichweite]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Raggiungere]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[到達]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[대공]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Захват]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Alcance]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_REACH" />
  </STATIC_ABILITY>"""),

	("Defender",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Defender]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Défenseur]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Defensor.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Verteidiger]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Difensore]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[防衛]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[수비태세]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Защитник]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Defensor]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_DEFENDER" />
  </STATIC_ABILITY>"""),

	("Shroud",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Shroud]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Linceul]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Velo.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Verhüllt]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Velo]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[被覆]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[은폐]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Пелена]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Manto]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_SHROUD" />
  </STATIC_ABILITY>"""),

	("Indestructible",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Indestructible]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Indestructible]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Indestructible.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Unzerstörbar]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Indistruttibile]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[破壊不能]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[무적]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Неразрушимость]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Indestrutível]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_INDESTRUCTIBLE" />
  </STATIC_ABILITY>"""),

	("Flying",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Flying]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Vol]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Vuela.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Fliegend]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Volare]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[飛行]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[비행]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Полет]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Voar]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_FLYING" />
  </STATIC_ABILITY>"""),

("Shadow",
	r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Shadow]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Distorsion]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Se desvanece.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Irrealität]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Ombra]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[シャドー]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[Shadow]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Тень]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Sombra]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_SHADOW" />
  </STATIC_ABILITY>"""),

("Flash",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Flash]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Flash]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Destello.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Aufblitzen]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Lampo]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[瞬速]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[섬광]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Миг]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Lampejo]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_FLASH" />
  </STATIC_ABILITY>"""),

	("Vigilance",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Vigilance]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Vigilance]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Vigilancia.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Wachsamkeit]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Cautela]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[警戒]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[경계]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Бдительность]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Vigilância]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_VIGILANCE" />
  </STATIC_ABILITY>"""),

	("Haste",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Haste]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Célérité]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Prisa.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Eile]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Rapidità]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[速攻]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[신속]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Ускорение]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Ímpeto]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_HASTE" />
  </STATIC_ABILITY>"""),

	("Fear",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Fear]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Peur]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Inspirar temor.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Verursacht Furcht]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Paura]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[畏怖]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[공포]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Страх]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Amedrontar]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_FEAR" />
  </STATIC_ABILITY>"""),

	("Plainswalk",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Plainswalk]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Traversée des plaines]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Cruza llanuras.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Ebenentarnung]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Passa-pianure]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[平地渡り]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[Plainswalk]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Знание Равнин]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Travessia de planície]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_PLAINSWALK" />
  </STATIC_ABILITY>"""),

	("Islandwalk",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Islandwalk]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Traversée des îles]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Cruza islas.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Inseltarnung]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Passa-Isole]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[島渡り]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[Islandwalk]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Знание Островов]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Travessia de ilha]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_ISLANDWALK" />
  </STATIC_ABILITY>"""),

	("Swampwalk",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Swampwalk]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Traversée des marais]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Cruza pantanos]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Sumpftarnung]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Passa-Paludi]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[沼渡り]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[늪잠입]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Знание Болот]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Travessia de pântano]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_SWAMPWALK" />
  </STATIC_ABILITY>"""),

	("Mountainwalk",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Mountainwalk]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Traversée des montagnes]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Cruza montañas.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Gebirgstarnung]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Passa-montagne]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[山渡り]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[Mountainwalk]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Знание Гор]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Travessia de Montanha]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_MOUNTAINWALK" />
  </STATIC_ABILITY>"""),

	("Forestwalk",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Forestwalk]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Traversée des forêts]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Cruza bosques.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Waldtarnung]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Passa-Foreste]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[森渡り]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[숲잠입]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Знание Лесов]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Travessia de floresta]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_FORESTWALK" />
  </STATIC_ABILITY>"""),

	("Hexproof",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Hexproof]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Défense talismanique]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Antimaleficio.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Fluchsicher]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Anti-malocchio]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[呪禁]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[방호]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Порчеустойчивость]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Resistência à magia]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_HEXPROOF" />
  </STATIC_ABILITY>"""),

	("Intimidate",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Intimidate]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Intimidation]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Intimidar.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Einschüchtern]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Intimidire]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[威嚇]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[위협]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Устрашение]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Intimidar]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_INTIMIDATE" />
  </STATIC_ABILITY>"""),

	("Battle cry",
r"""<TRIGGERED_ABILITY badge="BADGE_BATTLE_CRY">
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Battle cry]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Cri de guerre]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Grito de batalla.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Schlachtruf]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Grido di guerra]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[喊声]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[전투함성]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Боевой клич]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Grito de guerra]]></LOCALISED_TEXT>
    <TRIGGER value="ATTACKING" simple_qualifier="self" />
    <FILTER filter_id="0">
    local filter = ClearFilter()
    filter:Add( FE_IS_ATTACKING, true )
    filter:Add( FE_CARD_INSTANCE, OP_NOT, TriggerObject() )
    </FILTER>
    <CONTINUOUS_ACTION layer="7C" filter_id="0">
    if FilteredCard() ~= nil then
    	local characteristics = FilteredCard():GetCurrentCharacteristics()
    	if characteristics ~= nil then
    		characteristics:Power_Add( 1 )
    	end
    end
    </CONTINUOUS_ACTION>
    <DURATION simple_duration="UntilEOT" />
  </TRIGGERED_ABILITY>"""),

	("Infect",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Infect]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Infettare]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Infizieren]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Infection]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Infectar.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[感染]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[감염]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Инфекция]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Infectar]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_INFECT" />
  </STATIC_ABILITY>"""),

	("Wither",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Wither]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Flétrissure]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Debilitar.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Verdorren]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Avvizzire]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[萎縮]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[고사]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Увядание]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Murchar]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_WITHER" />
  </STATIC_ABILITY>"""),

	("Exalted",
r"""<TRIGGERED_ABILITY badge="BADGE_EXALTED">
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Exalted]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Exaltation]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Exaltado.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Edelmut]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Esaltato]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[賛美]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[고귀]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Повышение]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Exaltado]]></LOCALISED_TEXT>
    <TRIGGER value="ATTACKING_ALONE" simple_qualifier="objectyoucontrol" />
    <CONTINUOUS_ACTION layer="7C">
    if TriggerObject() ~= nil then
    	local characteristics = TriggerObject():GetCurrentCharacteristics()
    	characteristics:Power_Add(1)
    	characteristics:Toughness_Add(1)
    end
    </CONTINUOUS_ACTION>
    <DURATION simple_duration="UntilEOT" />
    <AUTO_SKIP no_trigger_object="1" />
  </TRIGGERED_ABILITY>"""),

	("Undying",
r"""<TRIGGERED_ABILITY badge="BADGE_UNDYING">
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Undying]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Survivance]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Resiliencia.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Unverwüstlich]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Immortale]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[不死]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[불사]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Нетленность]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Imortal]]></LOCALISED_TEXT>
    <TRIGGER value="ZONECHANGE_BEGIN" simple_qualifier="self" to_zone="ZONE_GRAVEYARD" from_zone="ZONE_BATTLEFIELD" />
    <INTERVENING_IF ignore_resolution_check="1">
    return TriggerObject():CountCounters( MTG():PlusOnePlusOneCounters() ) == 0
    </INTERVENING_IF>
    <RESOLUTION_TIME_ACTION>
    if TriggerObject() ~= nil then
    	TriggerObject():PutOntoBattlefield( TriggerObject():GetOwner() )
    	TriggerObject():AddCounters( MTG():PlusOnePlusOneCounters(), 1 )
    end
    </RESOLUTION_TIME_ACTION>
  </TRIGGERED_ABILITY>"""),

	("Changeling",
r"""<STATIC_ABILITY active_zone="ZONE_ANY">
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Changeling]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Changelin]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Cambiaformas.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Wandelwicht]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Cangiante]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[多相]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[자유변형]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Перевертыш]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Morfoloide]]></LOCALISED_TEXT>
    <CONTINUOUS_ACTION layer="4">
    if EffectSource() ~= nil then
    	EffectSource():GetCurrentCharacteristics():Bool_Set( CHARACTERISTIC_CHANGELING, 1 )
    end
    </CONTINUOUS_ACTION>
  </STATIC_ABILITY>"""),

	("Persist",
r"""<TRIGGERED_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Persist]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Persistance]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Persistir.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Beharrlichkeit]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Persistere]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[頑強]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[집요]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Упорство]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Persistir]]></LOCALISED_TEXT>
    <TRIGGER value="ZONECHANGE_BEGIN" simple_qualifier="self" to_zone="ZONE_GRAVEYARD" from_zone="ZONE_BATTLEFIELD" />
    <INTERVENING_IF ignore_resolution_check="1">
    return TriggerObject():CountCounters( MTG():MinusOneMinusOneCounters() ) == 0
    </INTERVENING_IF>
    <RESOLUTION_TIME_ACTION>
    if TriggerObject() ~= nil then
    	TriggerObject():PutOntoBattlefield( TriggerObject():GetOwner() )
    	TriggerObject():AddCounters( MTG():MinusOneMinusOneCounters(), 1 )
    end
    </RESOLUTION_TIME_ACTION>
  </TRIGGERED_ABILITY>"""),

	("Flanking",
r"""<STATIC_ABILITY>
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Flanking]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Débordement]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Flanquea.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Flankenangriff]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Aggirare]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[側面攻撃]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[협공]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Обход]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Flanquear]]></LOCALISED_TEXT>
    <INTRINSIC characteristic="CHARACTERISTIC_FLANKING" />
  </STATIC_ABILITY>
  <TRIGGERED_ABILITY badge="BADGE_FLANKING">
    <TRIGGER value="WAS_BLOCKED_BY_A_CREATURE" simple_qualifier="self">
    return SecondaryObject():GetCurrentCharacteristics():Bool_Get( CHARACTERISTIC_FLANKING ) == false
    </TRIGGER>
    <CONTINUOUS_ACTION layer="7C">
    local blocker = SecondaryObject()
    if blocker ~= nil then    
    	blocker:GetCurrentCharacteristics():Power_Add( -1 )
    	blocker:GetCurrentCharacteristics():Toughness_Add( -1 )
    end
    </CONTINUOUS_ACTION>
    <DURATION simple_duration="UntilEOT" />
  </TRIGGERED_ABILITY>"""),

	("Totem armor",
r"""<TRIGGERED_ABILITY replacement_query="1">
    <LOCALISED_TEXT LanguageCode="en-US"><![CDATA[Totem armor]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="fr-FR"><![CDATA[Armure totémique]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="es-ES"><![CDATA[Armadura tótem.]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="de-DE"><![CDATA[Totembeistand]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="it-IT"><![CDATA[Armatura totem]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="jp-JA"><![CDATA[族霊鎧]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ko-KR"><![CDATA[토템아머]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="ru-RU"><![CDATA[Тотемная броня]]></LOCALISED_TEXT>
    <LOCALISED_TEXT LanguageCode="pt-BR"><![CDATA[Armadura de totem]]></LOCALISED_TEXT>
    <TRIGGER value="DESTROYED" pre_trigger="1">
    if TriggerObject() == EffectSource():GetParent() then
    	local filter = ClearFilter()
    	filter:Add( FE_PARENT_CARD, OP_IS, TriggerObject() )
    	filter:Add( FE_CHARACTERISTIC, OP_HAS, CHARACTERISTIC_TOTEM_ARMOUR )
    	filter:EvaluateObjects()
    	if EffectSource() == filter:GetNthEvaluatedObject(0) then
    		MTG():OverrideEvent()
    		return true
    	end
    end
    return false
    </TRIGGER>
    <RESOLUTION_TIME_ACTION>
    local source = EffectSource()
    if source ~= nil then
    	local parent = source:GetParent()
    	if parent ~= nil then
    		parent:ClearDamage()
    	end
    	local filter = ClearFilter()
    	filter:Add( FE_PARENT_CARD, OP_IS, TriggerObject() )
    	filter:Add( FE_CHARACTERISTIC, OP_HAS, CHARACTERISTIC_TOTEM_ARMOUR )
    	local filter_count = filter:EvaluateObjects()
    	if filter_count &gt; 0 then
    		if filter_count == 1 then
    			source:Destroy()
    		else
    			TriggerObject():GetPlayer():ChooseItem( "CARD_QUERY_CHOOSE_ENCHANTMENT_TO_DESTROY", EffectDC():Make_Targets(0) )
    		end
    	end
    end
    </RESOLUTION_TIME_ACTION>
    <RESOLUTION_TIME_ACTION>
    local to_destroy = EffectDC():Get_Targets(0) and EffectDC():Get_Targets(0):Get_CardPtr(0)
    if to_destroy ~= nil then
    	to_destroy:Destroy()
    end
    </RESOLUTION_TIME_ACTION>
  </TRIGGERED_ABILITY>
  <STATIC_ABILITY>
    <CONTINUOUS_ACTION layer="6">
    if EffectSource() ~= nil then
    	EffectSource():GetCurrentCharacteristics():Bool_Set( CHARACTERISTIC_TOTEM_ARMOUR, 1 )
    end
    </CONTINUOUS_ACTION>
    <CONTINUOUS_ACTION layer="8">
    local source = EffectSource()
    if source ~= nil then
    	local parent = source:GetParent()
    	if parent ~= nil then
    		parent:GetCurrentCharacteristics():Badge_Set( BADGE_TOTEM_ARMOR )
    	end
    end
    </CONTINUOUS_ACTION>
  </STATIC_ABILITY>""")

])

dicSimpleTemplatesHelp = dict([
	("First strike", '<HELP title="MORE_INFO_BADGE_TITLE_0" body="MORE_INFO_BADGE_BODY_0" zone="ZONE_ANY" />'),
	("Trample", '<HELP title="MORE_INFO_BADGE_TITLE_1" body="MORE_INFO_BADGE_BODY_1" zone="ZONE_ANY" />'),
	("Double strike", '<HELP title="MORE_INFO_BADGE_TITLE_2" body="MORE_INFO_BADGE_BODY_2" zone="ZONE_ANY" />'),
	("Deathtouch", '<HELP title="MORE_INFO_BADGE_TITLE_3" body="MORE_INFO_BADGE_BODY_3" zone="ZONE_ANY" />'),
	("Lifelink", '<HELP title="MORE_INFO_BADGE_TITLE_4" body="MORE_INFO_BADGE_BODY_4" zone="ZONE_ANY" />'),
	("Reach", '<HELP title="MORE_INFO_BADGE_TITLE_5" body="MORE_INFO_BADGE_BODY_5" zone="ZONE_ANY" />'),
	("Defender", '<HELP title="MORE_INFO_BADGE_TITLE_6" body="MORE_INFO_BADGE_BODY_6" zone="ZONE_ANY" />'),
	("Shroud", '<HELP title="MORE_INFO_BADGE_TITLE_7" body="MORE_INFO_BADGE_BODY_7" zone="ZONE_ANY" />'),
	("Indestructible", '<HELP title="MORE_INFO_BADGE_TITLE_8" body="MORE_INFO_BADGE_BODY_8" zone="ZONE_ANY" />'),
	("Flying", '<HELP title="MORE_INFO_BADGE_TITLE_10" body="MORE_INFO_BADGE_BODY_10" zone="ZONE_ANY" />'),
	("Shadow", '<HELP title="MORE_INFO_BADGE_TITLE_11" body="MORE_INFO_BADGE_BODY_11" zone="ZONE_ANY" />'),
	("Flash", '<HELP title="MORE_INFO_BADGE_TITLE_12" body="MORE_INFO_BADGE_BODY_12" zone="ZONE_ANY" />'),
	("Vigilance", '<HELP title="MORE_INFO_BADGE_TITLE_13" body="MORE_INFO_BADGE_BODY_13" zone="ZONE_ANY" />'),
	("Haste", '<HELP title="MORE_INFO_BADGE_TITLE_14" body="MORE_INFO_BADGE_BODY_14" zone="ZONE_ANY" />'),
	("Fear", '<HELP title="MORE_INFO_BADGE_TITLE_15" body="MORE_INFO_BADGE_BODY_15" zone="ZONE_ANY" />'),
	("Plainswalk", '<HELP title="MORE_INFO_BADGE_TITLE_16" body="MORE_INFO_BADGE_BODY_16" zone="ZONE_ANY" />'),
	("Islandwalk", '<HELP title="MORE_INFO_BADGE_TITLE_17" body="MORE_INFO_BADGE_BODY_17" zone="ZONE_ANY" />'),
	("Swampwalk", '<HELP title="MORE_INFO_BADGE_TITLE_18" body="MORE_INFO_BADGE_BODY_18" zone="ZONE_ANY" />'),
	("Mountainwalk", '<HELP title="MORE_INFO_BADGE_TITLE_19" body="MORE_INFO_BADGE_BODY_19" zone="ZONE_ANY" />'),
	("Forestwalk", '<HELP title="MORE_INFO_BADGE_TITLE_20" body="MORE_INFO_BADGE_BODY_20" zone="ZONE_ANY" />'),
	("Hexproof", '<HELP title="MORE_INFO_BADGE_TITLE_21" body="MORE_INFO_BADGE_BODY_21" zone="ZONE_ANY" />'),
	("Intimidate", '<HELP title="MORE_INFO_BADGE_TITLE_24" body="MORE_INFO_BADGE_BODY_24" zone="ZONE_ANY" />'),
	("Battle cry", '<HELP title="MORE_INFO_BADGE_TITLE_26" body="MORE_INFO_BADGE_BODY_26" zone="ZONE_ANY" />'),
	("Infect", '<HELP title="MORE_INFO_BADGE_TITLE_31" body="MORE_INFO_BADGE_BODY_31" zone="ZONE_ANY" />'),
	("Wither", '<HELP title="MORE_INFO_BADGE_TITLE_32" body="MORE_INFO_BADGE_BODY_32" zone="ZONE_ANY" />'),
	("Exalted", '<HELP title="MORE_INFO_BADGE_TITLE_33" body="MORE_INFO_BADGE_BODY_33" zone="ZONE_ANY" />'),
	("Undying", '<HELP title="MORE_INFO_BADGE_TITLE_34" body="MORE_INFO_BADGE_BODY_34" zone="ZONE_ANY" />'),
	("Changeling", '<HELP title="MORE_INFO_BADGE_TITLE_35" body="MORE_INFO_BADGE_BODY_35" zone="ZONE_ANY" />'),
	("Persist", '<HELP title="MORE_INFO_TITLE_PERSIST" body="MORE_INFO_BODY_PERSIST" zone="ZONE_ANY" />'),
	("Flanking", '<HELP title="MORE_INFO_BADGE_TITLE_30" body="MORE_INFO_BADGE_BODY_30" zone="ZONE_ANY" />'),
	("Totem armor", '<HELP title="MORE_INFO_BADGE_TITLE_28" body="MORE_INFO_BADGE_BODY_28" zone="ZONE_ANY" />')
])

#--------------------------------------------------------------------------------------------
# ABILITY-TYPE-GUESSING TEMPLATES
# The following templates are used in order to guess what kind of ability needs to be added.
#--------------------------------------------------------------------------------------------
dicAbilityTypeTemplates = dict([
	(r'^[^“]+: ', 'ACTIVATED'),
	(r'(^When |^Whenever |^If |^At )', 'TRIGGERED'),
	(r'(^As an additional cost )', 'UTILITY')
])

#--------------------------------------------------------------------------------------------
# PATTERN TEMPLATES
# The following templates are placed in the card when a certain regex pattern has been found.
# In the template, @n@ will be substituted by the n-th group in the regex.
#--------------------------------------------------------------------------------------------

dicPatternHeaders = dict([
	(r'^Enchant creature$', '<ACTIVATED_ABILITY>'),
	(r'^Enchant creature\W', '<SPELL_ABILITY attach_definition="0">'),
	(r'^Flashback\W', '<UTILITY_ABILITY qualifier="Flashback" active_zone="ZONE_GRAVEYARD">'),
	(r'^Replicate\W', '<UTILITY_ABILITY qualifier="Replicate">')
])

dicPatternHeadersHelp = dict([
	(r'^Flashback\W', '<HELP title="MORE_INFO_TITLE_FLASHBACK" body="MORE_INFO_BODY_FLASHBACK" zone="ZONE_ANY" />'),
	(r'^Replicate\W', '<HELP title="MORE_INFO_TITLE_REPLICATE" body="MORE_INFO_BODY_REPLICATE" zone="ZONE_ANY" />')
])

dicPatternCosts = dict([
	('^' + SimpleCostPattern() + '(: |, )', '<COST mana_cost="@1@" type="Mana" />'),
	('^(Flashback|Replicate|Equip)\W' + SimpleCostPattern() + '$', '<COST mana_cost="@2@" type="Mana" />'),
	('(^|, ){T}(: |, )', '<COST type="TapSelf" />'),
	('(^|, ){q}(: |, )', '<COST type="UntapSelf" />'),

	('(^S|, s)acrifice a creature(: |, )',
r"""<COST type="Sacrifice" definition="0" compartment="1" query_tag="CARD_QUERY_CHOOSE_CREATURE_TO_SACRIFICE" item_count="1" />
    <COST_DEFINITION id="0">
    local filter = ClearFilter()
    filter:Add( FE_TYPE, OP_IS, CARD_TYPE_CREATURE )
    </COST_DEFINITION>"""),

	('(^D|, d)iscard a card(: |, )',
r"""<COST type="Discard" definition="0" compartment="1" query_tag="CARD_QUERY_CHOOSE_CARD_TO_DISCARD" item_count="1" />
    <COST_DEFINITION id="0">
    local filter = ClearFilter()
    filter:SetZone( ZONE_HAND, EffectController() )
    </COST_DEFINITION>""")

])

dicPatternTriggers = dict([

	(r'^At the beginning of each (upkeep|player’s upkeep),',
r"""<TRIGGER value="BEGINNING_OF_STEP">
    return MTG():GetStep() == STEP_UPKEEP
    </TRIGGER>"""),

	(r'^At the beginning of your upkeep,',
r"""<TRIGGER value="BEGINNING_OF_PLAYERS_STEP" simple_qualifier="controller">
    return MTG():GetStep() == STEP_UPKEEP
    </TRIGGER>""")

])

dicPatternEffects = dict([

	(r'^Enchant creature$',
r"""<TARGET tag="CARD_QUERY_CHOOSE_CREATURE_TO_ENCHANT" definition="0" compartment="0" count="1" />
    <TARGET_DEFINITION id="0">
    local filter = ClearFilter()
    filter:Add( FE_TYPE, OP_IS, CARD_TYPE_CREATURE )
    </TARGET_DEFINITION>
    <RESOLUTION_TIME_ACTION>
    local target = EffectDC():Get_Targets(0):Get_CardPtr(0)
    if target ~= nil and EffectSource() ~= nil then
    	EffectSource():Attach( target )
    end
    </RESOLUTION_TIME_ACTION>
    <AI_SIMPLIFIED_TARGETING compartment="0" hint="HINT_NEUTRAL" />"""),

	(r'^Equip\W',
r"""<AVAILABILITY sorcery_time="1" />
    <AI_AVAILABILITY type="restriction" restriction_type="equip" />
    <TARGET tag="CARD_QUERY_CHOOSE_CREATURE_TO_EQUIP" definition="0" compartment="0" count="1" />
    <TARGET_DEFINITION id="0">
    local filter = ClearFilter()
    filter:Add( FE_TYPE, OP_IS, CARD_TYPE_CREATURE )
    filter:Add( FE_CONTROLLER, OP_IS, EffectController() )
    </TARGET_DEFINITION>
    <RESOLUTION_TIME_ACTION>
    local target_card = EffectDC():Get_Targets(0):Get_CardPtr(0)
    if target_card ~= nil and EffectSource() ~= nil then
    	EffectSource():Attach( target_card )	
    end
    </RESOLUTION_TIME_ACTION>
    <AUTO_SKIP>
    return EffectDC():Get_Targets(0):Get_CardPtr(0) == nil
    </AUTO_SKIP>"""),

	(r': Regenerate ',
r"""<RESOLUTION_TIME_ACTION>
    if EffectSource() ~= nil then
    	EffectSource():GiveRegeneration()
    end
    </RESOLUTION_TIME_ACTION>
    <AI_AVAILABILITY window_step="declare_blockers" type="window" window_in_combat="1" />
    <AI_AVAILABILITY type="in_response" response_source="1" />
    <AI_AVAILABILITY type="restriction" restriction_type="no_regeneration_shield" />
    <AI_MODIFIERS combat_immunity="1" />""")

])

dicPatternEffectsHelp = dict([
	(r': Regenerate ', '<HELP title="MORE_INFO_TITLE_REGENERATE" body="MORE_INFO_BODY_REGENERATE" zone="ZONE_ANY" />')
])